/* Copyright (C) 2009-2010 Tomasz Ostrowski <tomasz.o.ostrowski at gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation.

 * Miniscope is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * If the integration of miniscope code requires the use of a license
 * other than GPL, an alternative license may be arranged.
 */

#include "BusPirateScope.h"
#include "..\miniscope\device\MiniscopeDevice.h"
#include <windows.h>
#include <stdio.h>
#include <assert.h>
#include <algorithm>

extern int iComPortNumber;

DWORD WINAPI ThreadProc(LPVOID threaddata)
{
    enum { SAMPLES_COUNT = 512 };
    short samples[SAMPLES_COUNT];
    unsigned char data[SAMPLES_COUNT * sizeof(samples[0])];
    bool remains = false;
    busPirateScope->log("RX thread created");
    while (busPirateScope->connected)
    {
        if (busPirateScope->bManualTriggerRequest)
        {
            busPirateScope->bManualTriggerRequest = false;
            busPirateScope->sample_count = 0;
            if (busPirateScope->callbackTrigger)
                busPirateScope->callbackTrigger();
            if (busPirateScope->sampling_frame != busPirateScope->upcoming_sampling_frame)
            {
                busPirateScope->sampling_frame = busPirateScope->upcoming_sampling_frame;
                busPirateScope->log("Sampling frame size is %d now", busPirateScope->sampling_frame);
            }
        }

        // get samples from device
        DWORD readed;
        bool status = ReadFile(busPirateScope->hCommDev, remains?(data+1):data,
            remains?(sizeof(data)-1):sizeof(data), &readed, NULL);
        if (status == false)
        {
            busPirateScope->connected = false;
			if (busPirateScope->connected && busPirateScope->callbackConnect)
				busPirateScope->callbackConnect(DEVICE_DISCONNECTED, (char*)"ReadFile failed, disconnected");
            busPirateScope->log("RX thread exit");
            return 0;
        }
        //busPirateScope->log("RX: %d", readed);
        int total_bytes = remains?(readed+1):readed;
        int total_samples = total_bytes/2;
        for (int i=0; i<total_samples; i++)
        {
            samples[i] = 256 * data[2*i] + data[2*i+1];
        }

        if (total_bytes%2)
        {
            remains = true;
            data[0] = data[readed-1];
        }
        else
        {
            remains = false;
        }

        if (busPirateScope->connected &&
            busPirateScope->callbackData && (int)busPirateScope->sample_count < busPirateScope->sampling_frame)
        {
            int samples_to_send =
                std::min(int(busPirateScope->sampling_frame - busPirateScope->sample_count), total_samples);
            busPirateScope->callbackData(samples, samples_to_send * sizeof(samples[0]));
            //busPirateScope->log("Thread: %d samples pushed", samples_to_send);
            busPirateScope->sample_count += samples_to_send;
        }
        else
        {
            // ignore data, wait for next trigger
        }
    }
    busPirateScope->log("RX thread exit");
    return 0;
}

BusPirateScope::BusPirateScope():
    sampling_frame(0),
    upcoming_sampling_frame(1024),
    bManualTriggerRequest(false),
    Thread(NULL),
    hCommDev(INVALID_HANDLE_VALUE)
{
    sample_count = upcoming_sampling_frame;
    callbackData = NULL;
    callbackLog = NULL;
    callbackConnect = NULL;

    OutputDebugStr("busPirateScope created");
}

BusPirateScope::~BusPirateScope()
{
    Disconnect();
}

void BusPirateScope::log(const char* const lpData, ...)
{
    static char buf[256]; //determines max message length
    va_list ap;
    va_start(ap, lpData);
    vsnprintf(buf, sizeof(buf), lpData, ap);
    va_end(ap);
    buf[sizeof(buf)-1] = 0;

    if (callbackLog)
        callbackLog(buf);
}

int BusPirateScope::Connect(void)
{
	if (hCommDev != INVALID_HANDLE_VALUE)
		CloseHandle(hCommDev);
	hCommDev = INVALID_HANDLE_VALUE;

    COMMTIMEOUTS ctmoNew, ctmoOld;
    char szComPort[32];
    sprintf(szComPort, "\\\\.\\COM%d", iComPortNumber);
    hCommDev = CreateFile(szComPort, GENERIC_READ | GENERIC_WRITE, 0,
                          NULL, OPEN_EXISTING, 0, 0);
    if (hCommDev != INVALID_HANDLE_VALUE)
    {
        //ustawienia portu
        if (SetupComm(hCommDev, 8192, 4096) == false)
        {
            log("SetupComm call failed!");
        }
        //ustawienia timeoutow
        GetCommTimeouts(hCommDev,&ctmoOld);
        //maximum time, in milliseconds, allowed to elapse between the arrival of two characters
        ctmoNew.ReadIntervalTimeout = 100;
        ctmoNew.ReadTotalTimeoutConstant = 40; //czas dodawany dla calej operacji
        ctmoNew.ReadTotalTimeoutMultiplier = 10;   //to jest mnozone przez liczbe bajtow dla operacji IO
        ctmoNew.WriteTotalTimeoutMultiplier = 0;
        ctmoNew.WriteTotalTimeoutConstant = 600;
        SetCommTimeouts(hCommDev, &ctmoNew);

        // -- ustawienia flag sterujacych DCB --
        dcb.DCBlength = sizeof(dcb);
        GetCommState(hCommDev, &dcb);
        dcb.BaudRate = CBR_115200;
        //jaka i czy jest parzystosc
        dcb.Parity = NOPARITY;
        //liczba bitow stopu
        dcb.StopBits = ONESTOPBIT;
        //wybor liczby bitow danych
        dcb.ByteSize = 8;
        SetCommState(hCommDev, &dcb);
        log( "Opened port %s.", szComPort );
        FlushFileBuffers(hCommDev);
        //try identify device
        log("Trying to enter bitbang mode...");
        const char* id = "BBIOx";   // "x" is protocol version, ignored when comparing
        int i=0;
        for (i=0; i<20; i++)
        {
            DWORD written, readed;
            unsigned char data = 0x00;
            WriteFile(hCommDev, &data, 1, &written, NULL);
            FlushFileBuffers(hCommDev);
            //Sleep(300);
            char buf[32];
            ReadFile(hCommDev, buf, strlen(id), &readed, NULL);
            log("Received %d bytes, [%.*s]", readed, readed, buf);
            if (readed == strlen(id))
            {
                PurgeComm(hCommDev, PURGE_TXCLEAR | PURGE_RXCLEAR);
                if (memcmp(buf, id, strlen(id)-1) == 0)
                {
                    unsigned char data = 0x15;
                    WriteFile(hCommDev, &data, 1, &written, NULL);
                    if (written == 1)
                    {
                        ConnectCb(DEVICE_CONNECTED, (char*)"Bus Pirate");
                        log("Entered bitbang mode!");
                        connected = true;
                        DWORD dwtid;
                        if (!Thread)
                            Thread = CreateThread(NULL, 0, ThreadProc, this, 0, &dwtid);
                        SetThreadPriority(Thread, THREAD_PRIORITY_TIME_CRITICAL);

                        return 0;
                    }
                }
            }
        }
        if (i==20)
        {
            ConnectCb(DEVICE_DISCONNECTED, (char*)"Failed to enter bitbang mode (not a Bus Pirate?).");
            log("Not a Bus Pirate at port %s? Change configuration (.cfg file) and restart.", szComPort);
            return 1;
        }
    }
    else
    {
        ConnectCb(DEVICE_DISCONNECTED, (char*)"Could not connect to port. Check if port number in .cfg file is correct and restart.");
        log("Failed opening %s. Change configuration (.cfg file) and restart.", szComPort);
        return 1;
    }

    ConnectCb(DEVICE_DISCONNECTED, (char*)"Failed to connect.");
    log("Failed to connect.");
    return 1;
}

int BusPirateScope::Disconnect(void)
{
#if 1
    // try to exit continuous sampling mode
    DWORD written;
    unsigned char data[2];
    memset(data, 0x0F, sizeof(data));
    WriteFile(hCommDev, &data, sizeof(data), &written, NULL);
    //FlushFileBuffers(hCommDev);
#endif
    // wait for read thread to terminate
    connected = false;
    Sleep(500);

	if (hCommDev != INVALID_HANDLE_VALUE)
		CloseHandle(hCommDev);

    Thread = NULL;

    return 0;
}

int BusPirateScope::ManualTrigger(void)
{
    bManualTriggerRequest = true;
    return 0;
}
