#ifndef BUSPIRATESCOPE_H_INCLUDED
#define BUSPIRATESCOPE_H_INCLUDED

#include "..\miniscope\device\MiniscopeDeviceCapabilities.h"
#include <vector>
#include <string>
#include <windows.h>
#include "singleton.h"

class BusPirateScope : public CSingleton<BusPirateScope>
{
private:
    friend class CSingleton<BusPirateScope>;
	BusPirateScope();
	~BusPirateScope();

    unsigned int sample_count;  // counter for samples sent to user
    int sampling_frame;
    int upcoming_sampling_frame;
    bool bManualTriggerRequest;
    HANDLE Thread;
    HANDLE hCommDev;
	DCB dcb;
	DWORD fdwEvtMask;

	typedef void (*CallbackLog)(char*);
	typedef void (*CallbackData)(short*, int);
	typedef void (*CallbackConnect)(int, char*);
	typedef void (*CallbackTrigger)(void);
	void log(const char* const lpData, ...);

    friend DWORD WINAPI ThreadProc(LPVOID data);

public:

	int Connect(void);
	int Disconnect(void);
	bool connected;

	CallbackLog callbackLog;
	CallbackData callbackData;
	CallbackConnect callbackConnect;
	CallbackTrigger callbackTrigger;
	void ConnectCb(int state, char* msg) {
		if (callbackConnect)
			callbackConnect(state, msg);
	}
	/** \brief Request manual triggering
		\return 0 on success
	*/
	int ManualTrigger(void);

	int SetBufferSize (int size) {
	    upcoming_sampling_frame = size;
	    log("Buffer (upcoming) size set to %d samples.", upcoming_sampling_frame);
	    return 0;
	}
};

#define busPirateScope BusPirateScope::Instance()

#endif // BUSPIRATESCOPE_H_INCLUDED
